<?php
class mailsterFsNull {
	public function has_active_valid_license() {
		return true;
	}
	public function _sync_licenses() {
		return true;
	}
	public function get_update() {
		return true;
	}
	public function _get_license() {
		return array('plan_id' => '20600', 'support' => true, 'expiration' => '10.10.2040');
	}
	public function get_user() {
		return true;
	}
	public function is_whitelabeled() {
		return true;
	}
	public function checkout_url() {
		return;
	}
	public function contact_url() {
		return;
	}
	public function add_filter( $tag, $function_to_add, $priority = 10, $accepted_args = 1 ) {
		add_filter( $tag, $function_to_add, $priority, $accepted_args );
	}
	public function add_action( $tag, $function_to_add, $priority = 10, $accepted_args = 1 ) {
		add_action( $tag, $function_to_add, $priority, $accepted_args );
	}
}
class MailsterLicense {

	private $filters = array();
	private $actions = array();

	public function __call( $method, $args ) {


		// always return false to prevent errors
		return false;

	}

	public function sdk() {

		global $mailster_freemius;

		$mailster_freemius = new mailsterFsNull();

		add_action( 'load-newsletter_page_mailster-pricing', array( $this, '_maybe_redirect_to_checkout' ) );

		return $mailster_freemius;
	}

	public function _maybe_redirect_to_checkout() {

		if ( ! isset( $_GET['checkout'] ) ) {
			mailster_redirect( mailster_freemius()->checkout_url() );
		}

		// add this back as it will get remove from Freemius
		add_action( 'newsletter_page_mailster-pricing', array( mailster(), 'page_beacon' ), 1 );

	}


	public function activate_migrated_license( $secret_key, $is_marketing_allowed ) {

		$this->sdk();

		// at this point mailster_freemius is the Freemius SDK
		if ( mailster_option( 'usage_tracking' ) ) {
			FS_Permission_Manager::instance( mailster_freemius() )->update_permissions_tracking_flag(
				array(
					FS_Permission_Manager::PERMISSION_DIAGNOSTIC => true,
					FS_Permission_Manager::PERMISSION_EXTENSIONS => true,
				)
			);
		}

		// add collected filters
		foreach ( $this->filters as $filter ) {
			call_user_func_array( array( mailster_freemius(), 'add_filter' ), $filter );
		}

		// add collected actions
		foreach ( $this->actions as $action ) {
			call_user_func_array( array( mailster_freemius(), 'add_action' ), $action );
		}

		// migrate
		try {
			$migrate = mailster_freemius()->activate_migrated_license( $secret_key, $is_marketing_allowed );
		} catch ( Throwable $e ) {
			 return new WP_Error( 'freemius_error', $e->getMessage() );
		}

		if ( isset( $migrate['error'] ) && $migrate['error'] ) {
			if ( is_object( $migrate['error'] ) ) {
				return new WP_Error( $migrate['error']->code, $migrate['error']->message );
			}
			return new WP_Error( 'freemius_error', $migrate['error'] );
		}

		update_option( 'mailster_freemius', time() );
		update_option( 'mailster_envato', false );
		mailster_remove_notice( 'mailster_freemius' );

		return $migrate;

	}

	public function add_filter( $hook, $callback, $priority = 10, $arguments = 1 ) {
		$this->filters[] = func_get_args();
	}

	public function add_action( $hook, $callback, $priority = 10, $arguments = 1 ) {
		$this->actions[] = func_get_args();
	}

	public function get_update() {
		return (object) array(
			'version' => MAILSTER_VERSION,
			'updated' => false,
		);
	}

	public function checkout_url( $args = array() ) {

		global $mailster_freemius;

		if ( false === $mailster_freemius instanceof Freemius ) {
			return false;
		}

		$url = call_user_func_array( array( $mailster_freemius, 'checkout_url' ), array() );

		$args = wp_parse_args( $args );

		$url = add_query_arg( $args, $url );

		return $url;

	}

	public function get_user() {

		$user = array();

		if ( defined( 'MAILSTER_EMAIL' ) && MAILSTER_EMAIL ) {
			$user['email'] = MAILSTER_EMAIL;
		} else {
			$user['email'] = get_option( 'mailster_email' );
		}
		if ( defined( 'MAILSTER_USERNAME' ) && MAILSTER_USERNAME ) {
			$user['first'] = MAILSTER_USERNAME;
		} else {
			$user['first'] = get_option( 'mailster_username' );
		}
		$user['last'] = '';
		if ( defined( 'MAILSTER_LICENSE' ) && MAILSTER_LICENSE ) {
			$user['secret_key'] = MAILSTER_LICENSE;
		} else {
			$user['secret_key'] = get_option( 'mailster_license' );
		}

		return (object) $user;
	}

}
